/*
 *  OpenGL-HQ rendering code Copyright (C) 2004-2005 Jörg Walter <jwalt@garni.ch>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
/* $Id$ */
/*---------------------------------------------------------------------
  This code implements a hardware-accelerated, cross-platform OpenGL
  based scaler quite similar to the well known Hq2x..Hq4x suite of
  software scalers. The general idea is exactly the same, but nothing
  else remains: in contrast to my software-hq2x scaler for dosbox, not
  even the interpolation rules have been used. Instead, they were
  designed from scratch to look well even on high scaling factors.

  Some compromises had to be taken, and it is indeed possible that
  my choice of rules don't fit some particular game. For this reason,
  the built-in table can be overridden by placing a file called
  "dosbox_openglhq_table.dat" into the working directory. It will
  be loaded instead of the shipped table. Use the tablebuilder to
  edit the table file. Likewise, you can change the fragment programs
  used by creating files called "dosbox_openglhq_passN.fp" for
  N = 1..3.

  As with the software hq2x scaler, the difference calculation is
  based on the algorithm described in
  http://www.compuphase.com/cmetric.htm, and the edge threshold
  calculation has entirely been conceived by me.

  Limitations:

  - None.

  - No really, there are none.

  - Okay, software Hq2x may be slightly faster. If 2x scaling w/o
    aspect ratio correction is all you need, try that and compare.

  - Everything said applies to a Mobility Radeon 9700. Faster cards
    might yield the previous statement untrue, I don't own an Nvidia
    card, etc. While I tried to create a universal high-performance,
    high-quality scaler, YMMV and Send Patches(tm).

*/

#ifndef _GNU_SOURCE
#define _GNU_SOURCE
#endif

#include "SDL.h"
#include "SDL_ohqvideo.h"

#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <unistd.h>
#include <stdarg.h>
#include <sys/types.h>
#include <math.h>
#include <fcntl.h>
#include <assert.h>

#include "SDL_opengl.h"

#ifdef WIN32
#define OPENGL_1_1
#endif
#ifndef APIENTRY
#define APIENTRY
#endif
#ifndef O_BINARY
#define O_BINARY (0)
#endif

#define CALL_ASYNC
#define RENDER_ASYNC

typedef void * (APIENTRY * OHQ_malloc_t) (int size, float readfreq, float writefreq, float priority);
typedef void (APIENTRY * OHQ_free_t) (void *pointer);

static OHQ_malloc_t OHQ_malloc = NULL;
static OHQ_free_t OHQ_free = NULL;

#ifdef OPENGL_1_1
static PFNGLACTIVETEXTUREPROC glActiveTexture = NULL;
static PFNGLTEXIMAGE3DPROC glTexImage3D = NULL;
#endif

static PFNGLPIXELDATARANGENVPROC glPixelDataRangeNV = NULL;

/* supported only on fairly recent video cards (somewhere around ATI Radeon 9500 or a similar NVidia card) */
static PFNGLPROGRAMSTRINGARBPROC glProgramStringARB = NULL;
static PFNGLBINDPROGRAMARBPROC glBindProgramARB = NULL;
static PFNGLGENPROGRAMSARBPROC glGenProgramsARB = NULL;
static PFNGLDELETEPROGRAMSARBPROC glDeleteProgramsARB = NULL;
static PFNGLGETPROGRAMIVARBPROC glGetProgramivARB = NULL;
static PFNGLPROGRAMLOCALPARAMETER4DARBPROC glProgramLocalParameter4dARB = NULL;
static PFNGLPROGRAMENVPARAMETER4DARBPROC glProgramEnvParameter4dARB = NULL;

#include "SDL_openglhq_pass1.h"
#include "SDL_openglhq_pass2.h"
#include "SDL_openglhq_pass3.h"
#include "SDL_openglhq_table.h"

#define OGL_HQ_RESOLUTION 16
/*#define OGL_DEBUG_HQ*/
/*#define OGL_DEBUG_HQ_MAX*/

#define fmax(x,y) ((x)>(y)?(x):(y))
#define fmin(x,y) ((x)<(y)?(x):(y))

static int safe_semwait(SDL_sem *sem) {
	while (SDL_SemWait(sem) != 0) SDL_Delay(1);
	return 0;
}
#define SDL_SemWait safe_semwait

static int int_log2 (int val) {
    int log = 0;
    while ((val >>= 1) != 0)
	log++;
    return log;
}

/* instead of a plain malloc, we try to be nice to DMA-using hardware */
#define MALLOC_ALIGN 4096
#define align_upwards(x) ((void *)((int)((x) + MALLOC_ALIGN - 1) & ~(MALLOC_ALIGN-1)))
static void * APIENTRY default_malloc(int size, float readfreq, float writefreq, float priority) {
	char *ptr = (char *)malloc(size+sizeof(void*)+MALLOC_ALIGN-1);
	char *retval = (char *)align_upwards(ptr + sizeof(void*));
	void **real_ptr = (void **)(retval-sizeof(void*));
	*real_ptr = ptr;
	return retval;
}

static void APIENTRY default_free(void *pointer) {
	void **real_ptr = (void **)((char *)pointer-sizeof(void*));
	free(*real_ptr);
}

static unsigned int LoadNativeFragmentProgram(const char *program, const char *filename) {
    GLuint name;
    int errorPos, isNative, programFd;
    char programString[65536] = "";
    const char *dir;

    glGenProgramsARB(1,&name);
    glBindProgramARB(GL_FRAGMENT_PROGRAM_ARB,name);
    if ((dir = getenv("SDL_OPENGLHQ_DATA")) != NULL) {
    	strcpy(programString,dir);
	strcat(programString,filename);
	strcat(programString,".fp");

	if ((programFd = open(programString,O_RDONLY)) >= 0) {
	    read(programFd,programString,sizeof(programString));
	    close(programFd);
	    program = programString;
	}
    }
    glProgramStringARB(GL_FRAGMENT_PROGRAM_ARB, GL_PROGRAM_FORMAT_ASCII_ARB, strlen(program), program);

    glGetIntegerv(GL_PROGRAM_ERROR_POSITION_ARB, &errorPos);
    glGetProgramivARB(GL_FRAGMENT_PROGRAM_ARB, GL_PROGRAM_UNDER_NATIVE_LIMITS_ARB, &isNative);
    if (errorPos >= 0) {
        SDL_SetError("OPENGLHQ: Failed to load fragment program: %s",glGetString(GL_PROGRAM_ERROR_STRING_ARB));
        glDeleteProgramsARB(1,&name);
        return 0;
    } else if (!isNative) {
        SDL_SetError("OPENGL: Hardware doesn't support this fragment program");
        glDeleteProgramsARB(1,&name);
        return 0;
    }
    return name;
}

static double sign(double a) {
    return (a < 0?-1:1);
}

/*
 This function calculates what percentage of a rectangle intersected by a line lies near the center of the
 cordinate system. It is mathematically exact, and well-tested for xcenter > 0 and ycenter > 0 (it's only
 used that way). It should be correct for other cases as well, but well... famous last words :)
*/
static double intersect_any(double xcenter, double ycenter, double xsize, double ysize, double yoffset, double gradient) {
	double g = fabs(gradient)*xsize/ysize;
	double o = -((yoffset-ycenter) + gradient*xcenter)/ysize*sign(ycenter)*sign(yoffset)-g*0.5+0.5;
	double yl = o, yr = o+g, xb = -o/g, xt = (1-o)/g;
	double area = 1.0;

	if (yl >= 1.0) xt = xb = area = 0.0;
	else if (yl > 0.0) {
		area = 1.0-yl;
		xb = 0.0;
	}
	else if (yr <= 0.0) yl = yr = area = 1.0;
	else yl = o+xb*g;

	if (xt <= 0.0) yr = yl = area = 0.0;
	else if (xt < 1.0) {
		area *= xt;
		yr = 1.0;
	}
	else if (xb >= 1.0) xb = xt = area = 1.0;
	else xt = (yr-o)/g;

	area -= (xt-xb)*(yr-yl)/2;

	return area;
}

static double intersect_h(double xcenter, double ycenter, double xsize, double ysize) {
	return fmax(0.0,fmin(1.0,(.55-fabs(xcenter)+xsize/2.0)/xsize));
}

static double intersect_any_h(double xcenter, double ycenter, double xsize, double ysize, double yoffset, double gradient) {
	double hinside = intersect_h(xcenter,ycenter,xsize,ysize);
	return hinside*hinside*intersect_any(xcenter,ycenter,xsize,ysize,yoffset,gradient);
}

static double intersect_v(double xcenter, double ycenter, double xsize, double ysize) {
	return fmax(0.0,fmin(1.0,(.55-fabs(ycenter)+ysize/2.0)/ysize));
}

static double intersect_any_v(double xcenter, double ycenter, double xsize, double ysize, double yoffset, double gradient) {
	double vinside = intersect_v(xcenter,ycenter,xsize,ysize);
	return vinside*vinside*intersect_any(xcenter,ycenter,xsize,ysize,yoffset,gradient);
}

static double intersect_hv(double xcenter, double ycenter, double xsize, double ysize) {
	double hinside = intersect_h(xcenter,ycenter,xsize,ysize);
	double vinside = intersect_v(xcenter,ycenter,xsize,ysize);
	return (1-hinside)*(1-vinside)+hinside*vinside;
}

/* FIXME: not sure if this is correct, but it is rare enough and most likely near enough. fixes welcome :) */
static double intersect_any_hv(double xcenter, double ycenter, double xsize, double ysize, double yoffset, double gradient) {
	double hvinside = intersect_hv(xcenter,ycenter,xsize,ysize);
	return hvinside*hvinside*intersect_any(xcenter,ycenter,xsize,ysize,yoffset,gradient);
}

static double intersect_hvd(double xcenter, double ycenter, double xsize, double ysize) {
	return intersect_h(xcenter,ycenter,xsize,ysize)*intersect_v(xcenter,ycenter,xsize,ysize);
}

static void setinterp(double xcenter, double ycenter, double percentage_inside, int i1, int i2, int i3, int o1, int o2, int o3, unsigned char *factors) {
	double d0, d1, d2, d3, percentage_outside, totaldistance_i, totaldistance_o;
	xcenter = fabs(xcenter);
	ycenter = fabs(ycenter);
	d0 = (1-xcenter)*(1-ycenter);
	d1 = xcenter*(1-ycenter);
	d2 = (1-xcenter)*ycenter;
	d3 = xcenter*ycenter;
	if (i1 && i2) i3 = 0;
	if (o1 && o2) o3 = 0;
	percentage_outside = 1.0-percentage_inside;
	totaldistance_i = d0+i1*d1+i2*d2+i3*d3;
	totaldistance_o = o1*d1+o2*d2+o3*d3+1e-12; /* +1e-12: prevent division by zero */

	factors[1] = (unsigned char)(((d1/totaldistance_i*percentage_inside*i1)+(d1/totaldistance_o*percentage_outside*o1))*255+.5);
	factors[2] = (unsigned char)(((d2/totaldistance_i*percentage_inside*i2)+(d2/totaldistance_o*percentage_outside*o2))*255+.5);
	factors[3] = (unsigned char)(((d3/totaldistance_i*percentage_inside*i3)+(d3/totaldistance_o*percentage_outside*o3))*255+.5);
	factors[0] = 255-factors[1]-factors[2]-factors[3];/*(unsigned char)((d0/totaldistance_i*percentage_inside)*255+.5);*/
}

/* Wanna have gcc fun? #define this as a macro, get a fast machine and go fetch a coffe or two. See how it is used to get an idea why.
   I aborted compilation after 5 minutes of CPU time on an Athlon64 3700+. */
static int swap_bits(int num, int bit1, int bit2) {
	return ((num & ~(bit1|bit2))|((num&bit1)?bit2:0)|((num&bit2)?bit1:0));
}

/*static int TrySendAsyncCommand(_THIS, enum OGL_CMD command) {
	if (this->hidden->busy && SDL_SemTryWait(this->hidden->render_thread_ack) != 0) return 0;
	this->hidden->busy = 1;
	this->hidden->cmd = command;
	SDL_SemPost(this->hidden->render_thread_signal);
	return 1;
}*/

static void WaitCommand(_THIS) {
	if (this->hidden->busy) SDL_SemWait(this->hidden->render_thread_ack);
	this->hidden->busy = 0;
}

#ifdef CALL_ASYNC
static void SendAsyncCommand(_THIS, enum OGL_CMD command) {
	if (this->hidden->busy) SDL_SemWait(this->hidden->render_thread_ack);
	this->hidden->busy = 1;
	this->hidden->cmd = command;
	SDL_SemPost(this->hidden->render_thread_signal);
}
#endif

static int SendSyncCommand(_THIS, enum OGL_CMD command) {
        int result;

	if (this->hidden->busy) SDL_SemWait(this->hidden->render_thread_ack);
	this->hidden->busy = 1;
	this->hidden->cmd = command;
	SDL_SemPost(this->hidden->render_thread_signal);
	SDL_SemWait(this->hidden->render_thread_ack);

	result = (this->hidden->status != OGL_ERROR);
	this->hidden->busy = 0;
	return result;
}

static void DeinitOpenGL(_THIS) {
	int i;

	if (this->hidden->pbuffer[0] != NULL) SDL_GL_FreeRenderTarget(this->hidden->pbuffer[0]);
	if (this->hidden->pbuffer[1] != NULL) SDL_GL_FreeRenderTarget(this->hidden->pbuffer[1]);
	this->hidden->pbuffer[0] = NULL;
	this->hidden->pbuffer[1] = NULL;

	for (i = 0; i < sizeof(this->hidden->program_name)/sizeof(this->hidden->program_name[0]); i++) {
		if (this->hidden->program_name[i] != 0) {
			//glDeleteProgramsARB(1,&this->hidden->program_name[i]);
			this->hidden->program_name[i] = 0;
		}
	}

	if (glIsList(this->hidden->pbuffer_displaylist)) glDeleteLists(this->hidden->pbuffer_displaylist, 1);

	if (this->hidden->framebufs[0]) {
		OHQ_free(this->hidden->framebufs[0]);
		this->hidden->framebufs[0] = NULL;
	}
	if (glIsList(this->hidden->displaylist)) glDeleteLists(this->hidden->displaylist, 1);

	if (this->hidden->texture[0]) {
		glDeleteTextures(4,this->hidden->texture);
		this->hidden->texture[0] = 0;
	}
}

static int RenderFrame(_THIS) {
	int x = this->hidden->clip.x, y = this->hidden->clip.y, cw = this->hidden->clip.w, ch = this->hidden->clip.h, tw = this->hidden->width, th = this->hidden->height;
	glTexSubImage2D(GL_TEXTURE_2D, 0, 0, 0,
			tw, th, this->hidden->framebuf_format,
			this->hidden->framebuf_datatype, this->hidden->framebufs[this->hidden->current_buffer]);
#ifdef RENDER_ASYNC
	SDL_SemPost(this->hidden->render_thread_ack);
#endif
#ifdef OGL_DEBUG_HQ_MAX
	GLubyte buffer[this->hidden->width*this->hidden->height*4];
	int fd = open("framebuf.pam",O_WRONLY|O_CREAT|O_TRUNC|O_BINARY,0666);
	sprintf((char *)buffer,"P7\nWIDTH %i\nHEIGHT %i\nMAXVAL 255\nDEPTH 4\nTUPLTYPE RGB_ALPHA\nENDHDR\n",this->hidden->width,this->hidden->height);
	write(fd,buffer,strlen((char *)buffer));
	write(fd,this->hidden->framebufs[0],this->hidden->pitch*this->hidden->height*2);
	close(fd);
#endif

	this->hidden->real_video->GL_LockRenderTarget(this->hidden->real_video,this->hidden->pbuffer[0]);
	/* It may seem inefficient and superfluous to do all that initialization again and again,
	   but it seems like render targets are a bit undefined in behaviour regarding these things.
	   Anyone with enough patience to double-check that some statements are not needed is welcome,
	   but honestly, I've tried quite a bit, and this seems to be a reasonable minimum which works
	   cross-platform.
	 */
	glViewport(x,th-y-ch,cw,ch);
	glMatrixMode (GL_MODELVIEW);
	glLoadIdentity ();
	glTranslated(-1,-1,0);
	glScaled(tw/(double)cw,th/(double)ch,1);
	glTranslated(1.0-(x/(double)tw*2),1.0-((th-y-ch)/(double)th*2),0);

	glActiveTexture(GL_TEXTURE0);
	glBindTexture(GL_TEXTURE_2D, this->hidden->texture[0]);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);

	glBindProgramARB(GL_FRAGMENT_PROGRAM_ARB,this->hidden->program_name[0]);
	glProgramLocalParameter4dARB(GL_FRAGMENT_PROGRAM_ARB,0,((double)this->hidden->static_threshold)/255.0,0.0,0.0,((double)this->hidden->dynamic_threshold)/100.0);
	glProgramEnvParameter4dARB(GL_FRAGMENT_PROGRAM_ARB,0,1.0/this->hidden->texsize,1.0/this->hidden->texsize,this->hidden->texsize,this->hidden->texsize);

	glEnable(GL_TEXTURE_2D);
	glEnable(GL_FRAGMENT_PROGRAM_ARB);

	glCallList(this->hidden->pbuffer_displaylist);
#ifdef OGL_DEBUG_HQ_MAX
	fd = open("diff1.pam",O_WRONLY|O_CREAT|O_TRUNC|O_BINARY,0666);
	sprintf((char *)buffer,"P7\nWIDTH %i\nHEIGHT %i\nMAXVAL 255\nDEPTH 4\nTUPLTYPE RGB_ALPHA\nENDHDR\n",this->hidden->width,this->hidden->height);
	write(fd,buffer,strlen((char *)buffer));
	glReadPixels(0,0,this->hidden->width,this->hidden->height,GL_RGBA,GL_UNSIGNED_BYTE,buffer);
	write(fd,buffer,sizeof(buffer));
	close(fd);
#endif
	this->hidden->real_video->GL_UnlockRenderTarget(this->hidden->real_video,this->hidden->pbuffer[0]);

	this->hidden->real_video->GL_LockRenderTarget(this->hidden->real_video,this->hidden->pbuffer[1]);
	glViewport(x,th-y-ch,cw,ch);
	glMatrixMode (GL_MODELVIEW);
	glLoadIdentity ();
	glTranslated(-1,-1,0);
	glScaled(tw/(double)cw,th/(double)ch,1);
	glTranslated(1.0-(x/(double)tw*2),1.0-((th-y-ch)/(double)th*2),0);

	glActiveTexture(GL_TEXTURE0);
	glBindTexture(GL_TEXTURE_2D,this->hidden->texture[1]);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);

	glBindProgramARB(GL_FRAGMENT_PROGRAM_ARB,this->hidden->program_name[1]);
	glProgramLocalParameter4dARB(GL_FRAGMENT_PROGRAM_ARB,0,((double)this->hidden->static_threshold)/255.0,0.0,0.0,((double)this->hidden->dynamic_threshold)/100.0);
	glProgramEnvParameter4dARB(GL_FRAGMENT_PROGRAM_ARB,0,1.0/this->hidden->texsize,1.0/this->hidden->texsize,this->hidden->texsize,this->hidden->texsize);

	glEnable(GL_TEXTURE_2D);
	glEnable(GL_FRAGMENT_PROGRAM_ARB);

	glCallList(this->hidden->pbuffer_displaylist);
#ifdef OGL_DEBUG_HQ_MAX
	fd = open("diff2.pam",O_WRONLY|O_CREAT|O_TRUNC|O_BINARY,0666);
	sprintf((char *)buffer,"P7\nWIDTH %i\nHEIGHT %i\nMAXVAL 255\nDEPTH 4\nTUPLTYPE RGB_ALPHA\nENDHDR\n",this->hidden->width,this->hidden->height);
	write(fd,buffer,strlen((char *)buffer));
	glReadPixels(0,0,this->hidden->width,this->hidden->height,GL_RGBA,GL_UNSIGNED_BYTE,buffer);
	write(fd,buffer,sizeof(buffer));
	close(fd);
#endif
	this->hidden->real_video->GL_UnlockRenderTarget(this->hidden->real_video,this->hidden->pbuffer[1]);

	glActiveTexture(GL_TEXTURE1);
	glBindTexture(GL_TEXTURE_2D,this->hidden->texture[2]);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
	glActiveTexture(GL_TEXTURE0);
	glBindTexture(GL_TEXTURE_2D,this->hidden->texture[0]);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);

	glBindProgramARB(GL_FRAGMENT_PROGRAM_ARB,this->hidden->program_name[2]);
	glProgramLocalParameter4dARB(GL_FRAGMENT_PROGRAM_ARB,0,((double)this->hidden->static_threshold)/255.0,0.0,0.0,((double)this->hidden->dynamic_threshold)/100.0);
	glProgramEnvParameter4dARB(GL_FRAGMENT_PROGRAM_ARB,0,1.0/this->hidden->texsize,1.0/this->hidden->texsize,this->hidden->texsize,this->hidden->texsize);
	if (x > 0) x++;
	if (y > 0) y++;
	if (cw < tw) cw-=2;
	if (ch < th) ch-=2;
	glViewport(x*this->hidden->outwidth/tw,y*this->hidden->outheight/th,cw*this->hidden->outwidth/tw,ch*this->hidden->outheight/th);
	glMatrixMode (GL_MODELVIEW);
	glLoadIdentity ();
	glTranslated(-1,-1,0);
	glScaled(tw/(double)cw,th/(double)ch,1);
	glTranslated(1.0-(x/(double)tw*2),1.0-(y/(double)th*2),0);

	glCallList(this->hidden->displaylist);
	this->hidden->real_video->GL_SwapBuffers(this->hidden->real_video);

	return 1;
}

static int InitOpenGL(_THIS) {
        int w, h, bpp;
	int border, x, y, texsize;
        const char *res;
	SDL_Surface *surface;
	GLfloat tex_width, tex_height;
	GLubyte *texture;
	double xsize, ysize;
	int attrib[] = {
		SDL_GL_RED_SIZE, 8,
		SDL_GL_GREEN_SIZE, 8,
		SDL_GL_BLUE_SIZE, 8,
		SDL_GL_ALPHA_SIZE, 8,
		SDL_GL_DOUBLEBUFFER, 0,
		0,
	};
	unsigned char table[4096] = dosbox_openglhq_table_dat;

	SDL_GL_SetAttribute(SDL_GL_DOUBLEBUFFER, 1);
	if (this->hidden->flags & SDL_FULLSCREEN) {
	   if ((res = getenv("SDL_OPENGLHQ_FULLRES")) != NULL) {
	       w = atoi(res);
	       res = strchr(res,'x');
	       if (res == NULL) {
	           SDL_SetError("Invalid mode specification in SDL_OPENGLHQ_FULLRES");
	           return 0;
	       }
	       h = atoi(res+1);
	       res = strchr(res,'-');
	       if (res != NULL) {
	           bpp = atoi(res+1);
	       } else {
	           bpp = this->hidden->format.BitsPerPixel;
               }
	   } else {
	       w = this->hidden->screen_resolution->w;
	       h = this->hidden->screen_resolution->h;
	       bpp = this->hidden->format.BitsPerPixel;
	   }
	} else {
	   if ((res = getenv("SDL_OPENGLHQ_WINRES")) != NULL) {
	       w = atoi(res);
	       res = strchr(res,'x');
	       if (res == NULL) {
	           SDL_SetError("Invalid mode specification in SDL_OPENGLHQ_WINRES");
	           return 0;
	       }
	       h = atoi(res+1);
	       res = strchr(res,'-');
	       if (res != NULL) {
	           bpp = atoi(res+1);
	       } else {
	           bpp = this->hidden->format.BitsPerPixel;
               }
	   } else {
	       w = this->hidden->width;
	       h = this->hidden->height;
	       bpp = this->hidden->format.BitsPerPixel;
	   }
	}

	surface = SDL_SetVideoMode(w,h,bpp,SDL_OPENGL|SDL_HWSURFACE|(this->hidden->flags&SDL_FULLSCREEN));
	if (surface == NULL) {
		SDL_SetError("OPENGLHQ: Can't open drawing surface, are you running in 16bpp(or higher) mode?");
		return 0;
	}
	this->hidden->outwidth = w;
	this->hidden->outheight = h;

	if (this->hidden->max_texsize == 0) {
		const char * gl_ext = (const char *)glGetString (GL_EXTENSIONS);

		if (getenv("SDL_OPENGLHQ_STATIC") != NULL) this->hidden->static_threshold = atoi(getenv("SDL_OPENGLHQ_STATIC"));
		else this->hidden->static_threshold = 10;
		if (this->hidden->static_threshold > 255) this->hidden->static_threshold = 255;
		if (this->hidden->static_threshold < 0) this->hidden->static_threshold = 0;
		if (getenv("SDL_OPENGLHQ_DYNAMIC") != NULL) this->hidden->dynamic_threshold = atoi(getenv("SDL_OPENGLHQ_DYNAMIC"));
		else this->hidden->dynamic_threshold = 33;
		if (this->hidden->dynamic_threshold > 100) this->hidden->dynamic_threshold = 100;
		if (this->hidden->dynamic_threshold < 0) this->hidden->dynamic_threshold = 0;

		glGetIntegerv(GL_MAX_TEXTURE_SIZE, &this->hidden->max_texsize);
		glPixelDataRangeNV = (PFNGLPIXELDATARANGENVPROC)SDL_GL_GetProcAddress("glPixelDataRangeNV");

		if (gl_ext != NULL && *gl_ext) {
			this->hidden->has_pixel_data_range=(strstr(gl_ext,"GL_NV_pixel_data_range") != NULL) && glPixelDataRangeNV != NULL;
		} else {
			this->hidden->has_pixel_data_range = 0;
		}

		if (this->hidden->has_pixel_data_range) {
			OHQ_malloc = (OHQ_malloc_t)SDL_GL_GetProcAddress("wglAllocateMemoryNV");
			if (OHQ_malloc == NULL) OHQ_malloc = (OHQ_malloc_t)SDL_GL_GetProcAddress("glXAllocateMemoryNV");
			OHQ_free = (OHQ_free_t)SDL_GL_GetProcAddress("wglFreeMemoryNV");
			if (OHQ_free == NULL) OHQ_free = (OHQ_free_t)SDL_GL_GetProcAddress("glXFreeMemoryNV");
		}
		if (OHQ_malloc == NULL) OHQ_malloc = (OHQ_malloc_t)default_malloc;
		if (OHQ_free == NULL) OHQ_free = (OHQ_free_t)default_free;

#ifdef OPENGL_1_1
		glActiveTexture = (PFNGLACTIVETEXTUREPROC)SDL_GL_GetProcAddress("glActiveTexture");
		glTexImage3D = (PFNGLTEXIMAGE3DPROC)SDL_GL_GetProcAddress("glTexImage3D");
#endif
		glProgramStringARB = (PFNGLPROGRAMSTRINGARBPROC)SDL_GL_GetProcAddress("glProgramStringARB");
		glBindProgramARB = (PFNGLBINDPROGRAMARBPROC)SDL_GL_GetProcAddress("glBindProgramARB");
		glGenProgramsARB = (PFNGLGENPROGRAMSARBPROC)SDL_GL_GetProcAddress("glGenProgramsARB");
		glDeleteProgramsARB = (PFNGLDELETEPROGRAMSARBPROC)SDL_GL_GetProcAddress("glDeleteProgramsARB");
		glGetProgramivARB = (PFNGLGETPROGRAMIVARBPROC)SDL_GL_GetProcAddress("glGetProgramivARB");
		glProgramLocalParameter4dARB = (PFNGLPROGRAMLOCALPARAMETER4DARBPROC)SDL_GL_GetProcAddress("glProgramLocalParameter4dARB");
		glProgramEnvParameter4dARB = (PFNGLPROGRAMENVPARAMETER4DARBPROC)SDL_GL_GetProcAddress("glProgramEnvParameter4dARB");

		if (!gl_ext || !*gl_ext || !((strstr(gl_ext,"ARB_fragment_program") != NULL) &&
				glProgramStringARB && glBindProgramARB && glGenProgramsARB && glDeleteProgramsARB &&
				glGetProgramivARB && glProgramLocalParameter4dARB && glProgramEnvParameter4dARB)) {
			SDL_SetError("OPENGLHQ: Video driver doesn't support fragment programs");
			return 0;
		}

	}

	this->hidden->framebuf_format = GL_BGRA_EXT;
	this->hidden->framebuf_internalformat = GL_RGBA;
	this->hidden->framebuf_datatype = GL_UNSIGNED_BYTE;
	this->hidden->bpp = 32;
	if (this->hidden->bpp == 8) {
		this->hidden->framebuf_format = GL_COLOR_INDEX;
	} else if (this->hidden->bpp == 15) {
		this->hidden->framebuf_format = GL_RGB;
		this->hidden->framebuf_datatype = GL_UNSIGNED_SHORT_5_5_5_1;
	} else if (this->hidden->bpp == 16) {
		this->hidden->framebuf_format = GL_RGB;
		this->hidden->framebuf_datatype = GL_UNSIGNED_SHORT_5_6_5;
	} else {
		this->hidden->bpp = 32;
	}

	texsize=this->hidden->texsize=2 << int_log2(this->hidden->width > this->hidden->height ? this->hidden->width : this->hidden->height);
	tex_width=((GLfloat)(this->hidden->width)/(GLfloat)texsize);
	tex_height=((GLfloat)(this->hidden->height)/(GLfloat)texsize);

	if (texsize>this->hidden->max_texsize) {
		SDL_SetError("OPENGLHQ: No support for texturesize of %d, falling back to surface",texsize);
		return 0;
	}

	/* Create the texture and display list */
	glMatrixMode(GL_PROJECTION);
	glGenTextures(4,this->hidden->texture);

	this->hidden->pitch=this->hidden->width*(this->hidden->bpp == 32?4:this->hidden->bpp == 8?1:2);
	this->hidden->framebufs[0] = OHQ_malloc(this->hidden->pitch*this->hidden->height*2+MALLOC_ALIGN,0.0,1.0,1.0);;
	this->hidden->framebufs[1] = align_upwards(((char *)this->hidden->framebufs[0])+this->hidden->pitch*this->hidden->height);

	if (this->hidden->has_pixel_data_range) {
		glPixelDataRangeNV(GL_WRITE_PIXEL_DATA_RANGE_NV,this->hidden->width*this->hidden->height*4*2,this->hidden->framebufs[0]);
		glEnableClientState(GL_WRITE_PIXEL_DATA_RANGE_NV);
	}

	if ((this->hidden->program_name[0] = LoadNativeFragmentProgram(dosbox_openglhq_pass1_fp,"openglhq_pass1")) == 0 ||
		(this->hidden->program_name[1] = LoadNativeFragmentProgram(dosbox_openglhq_pass2_fp,"openglhq_pass2")) == 0 ||
		(this->hidden->program_name[2] = LoadNativeFragmentProgram(dosbox_openglhq_pass3_fp,"openglhq_pass3")) == 0) {
		SDL_SetError("OPENGLHQ: Hardware doesn't support this output");
		return 0;
	}

	glBindTexture(GL_TEXTURE_2D, this->hidden->texture[1]);
	this->hidden->pbuffer[0] = SDL_GL_CreateRenderTarget(texsize,texsize,attrib);
	if (this->hidden->pbuffer[0] != NULL) SDL_GL_BindRenderTarget(this->hidden->pbuffer[0], this->hidden->texture[1]);

	glBindTexture(GL_TEXTURE_2D, this->hidden->texture[2]);
	this->hidden->pbuffer[1] = SDL_GL_CreateRenderTarget(texsize,texsize,attrib);
	if (this->hidden->pbuffer[1] != NULL) SDL_GL_BindRenderTarget(this->hidden->pbuffer[1], this->hidden->texture[2]);

	if (this->hidden->pbuffer[0] == NULL || this->hidden->pbuffer[1] == NULL) {
		SDL_SetError("OPENGLHQ: Could not allocate render buffer.");
		return 0;
	}

	glActiveTexture(GL_TEXTURE2);
	glEnable(GL_TEXTURE_3D);
	glBindTexture(GL_TEXTURE_3D, this->hidden->texture[3]);
	glTexEnvi(GL_TEXTURE_ENV,GL_TEXTURE_ENV_MODE,GL_COMBINE);
	glTexEnvi(GL_TEXTURE_ENV,GL_COMBINE_RGB,GL_REPLACE);
	glTexEnvi(GL_TEXTURE_ENV,GL_COMBINE_ALPHA,GL_REPLACE);
	glTexEnvi(GL_TEXTURE_ENV,GL_SRC0_RGB,GL_TEXTURE0);
	glTexEnvi(GL_TEXTURE_ENV,GL_OPERAND0_RGB,GL_SRC_COLOR);
	glTexEnvi(GL_TEXTURE_ENV,GL_SRC0_ALPHA,GL_TEXTURE0);
	glTexEnvi(GL_TEXTURE_ENV,GL_OPERAND0_ALPHA,GL_SRC_ALPHA);

	texture = (GLubyte *)malloc(OGL_HQ_RESOLUTION*OGL_HQ_RESOLUTION*4096*4);
	xsize = (double)this->hidden->width/(double)this->hidden->outwidth;
	ysize = (double)this->hidden->height/(double)this->hidden->outheight;

	/*
	Table layout is a bit convoluted to be better processable in the
	fragment programs and to save space. Use the table builder
	to modify it. If you are interested in details, read it's source.
	*/
	if (getenv("SDL_OPENGLHQ_DATA")) {
		char file[4096];
		int tablefd;
		strcpy(file,getenv("SDL_OPENGLHQ_DATA"));
		strcat(file,"openglhq_table.dat");
		if ((tablefd = open(file,O_RDONLY|O_BINARY)) >= 0) {
			read(tablefd,table,sizeof(table));
			close(tablefd);
		}
	}

#define R 1
#define T 2
#define RT 4
#define RT2 8
#define L 16
#define LB2 32
#define LT2 64
#define LT 128
#define LB 256
#define B 512
#define RB2 1024
#define RB 2048

#define NODIAG 0x90
#define H 1
#define V 2
#define D 4

#define hmirror(p) swap_bits(swap_bits(swap_bits(swap_bits(swap_bits(p,R,L),RT,LT),RT2,LT2),RB,LB),RB2,LB2)
#define vmirror(p) swap_bits(swap_bits(swap_bits(swap_bits(swap_bits(p,T,B),RT,RB),RT2,RB2),LT,LB),LT2,LB2)
#define NO_BORDER(x) ((b&(x)) == 0)
#define IS_BORDER(x) ((b&(x)) == (x))
#define SETINTERP(percentage_inside) setinterp(xcenter,ycenter,percentage_inside, \
						NO_BORDER(R),NO_BORDER(T),NO_BORDER(RT), \
						IS_BORDER(R),IS_BORDER(T),IS_BORDER(RT), \
						texture+((x+(border%16)*OGL_HQ_RESOLUTION+y*16*OGL_HQ_RESOLUTION+(border&~15)*OGL_HQ_RESOLUTION*OGL_HQ_RESOLUTION)*4))

	for (border = 0; border < 4096; border++) {
		for (y = 0; y < OGL_HQ_RESOLUTION; y++) {
			for (x = 0; x < OGL_HQ_RESOLUTION; x++) {
				double xcenter = fabs((((double)x)+.5)/(double)(OGL_HQ_RESOLUTION)-.5)/.958;
				double ycenter = fabs((((double)y)+.5)/(double)(OGL_HQ_RESOLUTION)-.5)/.958;
				int sx = (x < OGL_HQ_RESOLUTION/2?-1:1);
				int sy = (y < OGL_HQ_RESOLUTION/2?-1:1);
				int b = (sy > 0?(sx > 0?border:hmirror(border)):(sx > 0?vmirror(border):vmirror(hmirror(border))));

				if ((table[b] & NODIAG) == NODIAG) {
					if (table[b] & H) {
						if (table[b] & V) {
							if (table[b] & D) SETINTERP(intersect_hvd(xcenter,ycenter,xsize,ysize));
							else SETINTERP(intersect_hv(xcenter,ycenter,xsize,ysize));
						} else {
							SETINTERP(intersect_h(xcenter,ycenter,xsize,ysize));
						}
					} else if (table[b] & V) {
						SETINTERP(intersect_v(xcenter,ycenter,xsize,ysize));
					} else {
						SETINTERP(1.0);
					}
				} else {
					double yoff = (table[b]&4?1:-1)*(((table[b] >> 3) & 3) + 1)/4.0;
					double grad = (table[b]&32?1:-1)*(((table[b] >> 6) & 3) + 1)/2.0;
					if (table[b] & H) {
						if (table[b] & V) {
							SETINTERP(intersect_any_hv(xcenter,ycenter,xsize,ysize,yoff,grad));
						} else {
							SETINTERP(intersect_any_h(xcenter,ycenter,xsize,ysize,yoff,grad));
						}
					} else if (table[b] & V) {
						SETINTERP(intersect_any_v(xcenter,ycenter,xsize,ysize,yoff,grad));
					} else {
						SETINTERP(intersect_any(xcenter,ycenter,xsize,ysize,yoff,grad));
					}
				}

			}
		}
	}
#ifdef OGL_DEBUG_HQ_MAX
	int fd = open("texture.pam",O_WRONLY|O_CREAT|O_TRUNC|O_BINARY,0666);
	sprintf((char *)table,"P7\nWIDTH %i\nHEIGHT %i\nMAXVAL 255\nDEPTH 4\nTUPLTYPE RGB_ALPHA\nENDHDR\n",16*OGL_HQ_RESOLUTION,4096/16*OGL_HQ_RESOLUTION);
	write(fd,table,strlen((char *)table));
	write(fd,texture,sizeof(texture));
	close(fd);
#endif

	glTexImage3D(GL_TEXTURE_3D, 0, GL_RGBA8, 256, 16, 256, 0, GL_RGBA, GL_UNSIGNED_BYTE, texture);
	free(texture);
	glTexParameteri(GL_TEXTURE_3D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
	glTexParameteri(GL_TEXTURE_3D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
	glTexParameteri(GL_TEXTURE_3D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
	glTexParameteri(GL_TEXTURE_3D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
	glTexParameteri(GL_TEXTURE_3D, GL_TEXTURE_WRAP_R, GL_CLAMP_TO_EDGE);
	glTexEnvi(GL_TEXTURE_ENV,GL_TEXTURE_ENV_MODE,GL_COMBINE);
	glTexEnvi(GL_TEXTURE_ENV,GL_COMBINE_RGB,GL_REPLACE);
	glTexEnvi(GL_TEXTURE_ENV,GL_COMBINE_ALPHA,GL_REPLACE);
	glTexEnvi(GL_TEXTURE_ENV,GL_SRC0_RGB,GL_TEXTURE0);
	glTexEnvi(GL_TEXTURE_ENV,GL_OPERAND0_RGB,GL_SRC_COLOR);
	glTexEnvi(GL_TEXTURE_ENV,GL_SRC0_ALPHA,GL_TEXTURE0);
	glTexEnvi(GL_TEXTURE_ENV,GL_OPERAND0_ALPHA,GL_SRC_ALPHA);
	glActiveTexture(GL_TEXTURE0);

	this->hidden->pbuffer_displaylist = glGenLists(1);
	glNewList(this->hidden->pbuffer_displaylist, GL_COMPILE);
	glBegin(GL_QUADS);
	glTexCoord2f(0,0); glVertex2f(-1.0f,-1.0f);
	glTexCoord2f(tex_width,0); glVertex2f(1.0f, -1.0f);
	glTexCoord2f(tex_width,tex_height); glVertex2f(1.0f, 1.0f);
	glTexCoord2f(0,tex_height); glVertex2f(-1.0f, 1.0f);
	glEnd();
	glEndList();

	glEnable(GL_FRAGMENT_PROGRAM_ARB);

	glBindTexture(GL_TEXTURE_2D,this->hidden->texture[0]);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);

	glTexImage2D(GL_TEXTURE_2D, 0, this->hidden->framebuf_internalformat, texsize, texsize, 0, this->hidden->framebuf_format, this->hidden->framebuf_datatype, 0);
	glClearColor (0.0, 0.0, 0.0, 1.0);
	glClear(GL_COLOR_BUFFER_BIT);
	SDL_GL_SwapBuffers();
	glClear(GL_COLOR_BUFFER_BIT);
	glShadeModel (GL_FLAT);
	glDisable(GL_DEPTH_TEST);
	glDisable(GL_ALPHA_TEST);
	glDisable(GL_STENCIL_TEST);
	glDisable(GL_SCISSOR_TEST);
	glDisable(GL_BLEND);
	glDisable(GL_DITHER);
	glDisable(GL_INDEX_LOGIC_OP);
	glDisable(GL_COLOR_LOGIC_OP);
	glDisable(GL_LIGHTING);
	glDisable(GL_CULL_FACE);
	glDisable(GL_FOG);
	glEnable(GL_TEXTURE_2D);
	glMatrixMode (GL_MODELVIEW);
	glLoadIdentity ();

	this->hidden->displaylist = glGenLists(1);
	glNewList(this->hidden->displaylist, GL_COMPILE);
	glBindTexture(GL_TEXTURE_2D, this->hidden->texture[0]);
	glBegin(GL_QUADS);
	glTexCoord2f(0,tex_height); glVertex2f(-1.0f,-1.0f);
	glTexCoord2f(tex_width,tex_height); glVertex2f(1.0f, -1.0f);
	glTexCoord2f(tex_width,0); glVertex2f(1.0f, 1.0f);
	glTexCoord2f(0,0); glVertex2f(-1.0f, 1.0f);
	glEnd();
	glEndList();
	return 1;
}

static int RenderThread(void *data) {
	SDL_VideoDevice *this = (SDL_VideoDevice *)data;
	SDL_Rect **modes;
	this->hidden->real_video->VideoInit(this->hidden->real_video,(SDL_PixelFormat *)&this->hidden->format);
	this->hidden->real_video->GL_LoadLibrary(this->hidden->real_video,NULL);
	modes = SDL_ListModes(NULL,SDL_OPENGL|SDL_FULLSCREEN);
	if (((int)modes) != -1 && modes && *modes) this->hidden->screen_resolution = modes[0];
	SDL_SemPost(this->hidden->render_thread_ack);

	while (SDL_SemWait(this->hidden->render_thread_signal) == 0) {
		int ack = 0;
		this->hidden->status = OGL_RUNNING;
		if (this->hidden->cmd == OGL_CALL) {
			switch (this->hidden->call.type) {
			case OGL_CALL_P:
				this->hidden->call.func.p(this->hidden->real_video);
				break;
			case OGL_CALL_PII:
				this->hidden->call.func.pii(this->hidden->real_video,(int)this->hidden->call.args[0],(int)this->hidden->call.args[1]);
				break;
			case OGL_CALL_PPP:
				this->hidden->call.func.ppp(this->hidden->real_video,(const void*)this->hidden->call.args[0],(const void*)this->hidden->call.args[1]);
				break;
			case OGL_CALL_PP:
				this->hidden->call.func.pp(this->hidden->real_video,(const void*)this->hidden->call.args[0]);
				break;
			case OGL_CALL_PPPIIII_P:
				this->hidden->call.result = (intptr_t)this->hidden->call.func.pppiiii_p(this->hidden->real_video,(const void*)this->hidden->call.args[0],(const void*)this->hidden->call.args[1],
                                    (int)this->hidden->call.args[2],(int)this->hidden->call.args[3],(int)this->hidden->call.args[4],(int)this->hidden->call.args[5]);
				break;
			case OGL_CALL_PP_I:
				this->hidden->call.result = (intptr_t)this->hidden->call.func.pp_i(this->hidden->real_video,(const void*)this->hidden->call.args[0]);
				break;
			case OGL_CALL_P_I:
				this->hidden->call.result = (intptr_t)this->hidden->call.func.p_i(this->hidden->real_video);
				break;
			case OGL_CALL_PI_I:
				this->hidden->call.result = (intptr_t)this->hidden->call.func.pi_i(this->hidden->real_video,(int)this->hidden->call.args[0]);
				break;
			default:
				this->hidden->status = OGL_ERROR;
			}

		} else if (this->hidden->cmd == OGL_INIT) {
			if (!InitOpenGL(this)) {
				this->hidden->status = OGL_ERROR;
				DeinitOpenGL(this);
			}

		} else if (this->hidden->cmd == OGL_DEINIT) {
			DeinitOpenGL(this);

		} else if (this->hidden->cmd == OGL_FRAME) {
			if (!RenderFrame(this)) {
				this->hidden->status = OGL_ERROR;
#ifdef RENDER_ASYNC
			} else {
				ack = 1;
#endif
			}

		} else if (this->hidden->cmd == OGL_PALETTE) {
			glColorTable(GL_COLOR_TABLE,GL_RGBA,256,GL_RGBA8,GL_UNSIGNED_BYTE,this->hidden->pal);

		} else if (this->hidden->cmd == OGL_QUIT) {
			this->hidden->status = OGL_DONE;
			break;
		} else {
			this->hidden->status = OGL_ERROR;
		}
		if (this->hidden->status != OGL_ERROR) this->hidden->status = OGL_DONE;
		if (!ack) SDL_SemPost(this->hidden->render_thread_ack);
	}
	this->hidden->real_video->VideoQuit(this->hidden->real_video);
	SDL_SemPost(this->hidden->render_thread_ack);
	return 0;
}

static void StartRenderThread(_THIS) {
	this->hidden->cmd = OGL_DONE;
	this->hidden->render_thread_signal = SDL_CreateSemaphore(0);
	this->hidden->render_thread_ack = SDL_CreateSemaphore(0);
	this->hidden->render_thread = SDL_CreateThread(&RenderThread,this);
	SDL_SemWait(this->hidden->render_thread_ack);
}

static void ShutdownRenderThread(_THIS) {
    if (this->hidden->render_thread != NULL) {
        SendSyncCommand(this,OGL_DEINIT);
        SendSyncCommand(this,OGL_QUIT);
        SDL_WaitThread(this->hidden->render_thread, NULL);
        SDL_DestroySemaphore(this->hidden->render_thread_signal);
        SDL_DestroySemaphore(this->hidden->render_thread_ack);
        this->hidden->render_thread = NULL;
    }
}

